package jp.deadend.noname.howmm

import androidx.appcompat.widget.SearchView
import android.app.SearchManager
import android.appwidget.AppWidgetManager
import android.content.ComponentName
import android.content.Context
import android.content.Intent
import android.os.Bundle
import android.util.Log
import androidx.appcompat.app.AppCompatActivity
import android.view.KeyEvent
import android.view.Menu
import android.view.MenuItem
import android.view.View
import jp.deadend.noname.howmm.databinding.ActivityMainBinding
import kotlinx.coroutines.*
import java.io.IOException

class HowmmMain : AppCompatActivity() {
    private lateinit var binding: ActivityMainBinding
    private lateinit var mViewerPopup: TextViewerPopup
    private val mainScope = MainScope()

    public override fun onCreate(savedInstanceState: Bundle?) {
        super.onCreate(savedInstanceState)
        binding = ActivityMainBinding.inflate(layoutInflater)
        setContentView(binding.root)

        title = getString(R.string.app_name)

        binding.grepViewReminders.setCallback(object : GrepView.Callback {
            override fun onGrepItemClicked(position: Int) {
                val adapter = binding.grepViewReminders.adapter as GrepView.ReminderAdapter
                adapter.getItem(position)?.let { confirmDoneReminder(adapter, it) }
            }

            override fun onGrepItemLongClicked(position: Int): Boolean {
                editFile(binding.grepViewReminders.adapter.getItem(position) as GrepItem)
                return true
            }
        })
        binding.grepViewReminders.enableScroll(false)
        binding.grepViewRecent.setCallback(object : GrepView.Callback {
            override fun onGrepItemClicked(position: Int) {
                viewFile(binding.grepViewRecent.adapter.getItem(position) as GrepItem)
            }

            override fun onGrepItemLongClicked(position: Int): Boolean {
                editFile(binding.grepViewRecent.adapter.getItem(position) as GrepItem)
                return true
            }
        })
        binding.grepViewRecent.enableScroll(false)

        binding.floatingActionButton.setOnClickListener {
            if (mViewerPopup.isShowing) { mViewerPopup.dismiss() }
            val intent = Intent(this@HowmmMain, TextEditor::class.java).apply {
                putExtra(TextEditor.KEY_PATHNAME, null as String?)
                putExtra(TextEditor.KEY_TITLE, "")
            }
            startActivity(intent)
        }
        binding.textViewShowAll.setOnClickListener { doSearch("") }
    }

    override fun onCreateOptionsMenu(menu: Menu): Boolean {
        menuInflater.inflate(R.menu.menu_main, menu)
        val searchView = menu.findItem(R.id.menu_main_search).actionView as SearchView
        searchView.setOnQueryTextListener(object : SearchView.OnQueryTextListener {
            override fun onQueryTextSubmit(query: String): Boolean {
                doSearch(query.trim())
                return true
            }

            override fun onQueryTextChange(s: String) = false
        })
        searchView.setOnQueryTextFocusChangeListener { _, hasFocus ->
            if (hasFocus && mViewerPopup.isShowing) { mViewerPopup.dismiss() }
        }

        return super.onCreateOptionsMenu(menu)
    }

    override fun onOptionsItemSelected(item: MenuItem): Boolean {
        when (item.itemId) {
            R.id.menu_main_prefs -> {
                if (mViewerPopup.isShowing) { mViewerPopup.dismiss() }
                val intent = Intent(this@HowmmMain, SettingsActivity::class.java)
                startActivity(intent)
            }
            R.id.menu_main_search -> { }
            R.id.menu_main_hashtag -> chooseHashtag()
            R.id.menu_main_reload_reminder -> {
                clearCacheFile(TITLE_CACHE_FILE)
                clearCacheFile(REMINDER_CACHE_FILE)
                updateReminders()
                updateTitles()
            }
            else -> return super.onOptionsItemSelected(item)
        }

        return true
    }

    override fun onResume() {
        updateReminders()
        updateTitles()

        super.onResume()
    }

    override fun onAttachedToWindow() {
        mViewerPopup = TextViewerPopup(this, null)
        super.onAttachedToWindow()
    }

    private fun updateTitles() = mainScope.launch(Dispatchers.Main) {
        binding.progressBarRecent.visibility = View.VISIBLE
        binding.grepViewRecent.visibility = View.INVISIBLE
        val titles = grepTitles().sortedByDescending { it.file.lastModified() }
        if (titles.isNotEmpty()) {
            val adapter = GrepView.GrepAdapter(this@HowmmMain, titles)
            adapter.setFormat(null, howmmPrefs.fontSize)
            binding.grepViewRecent.adapter = adapter

            binding.textViewShowAll.visibility = View.GONE
            val listItem = adapter.getView(0, null, binding.grepViewRecent)
            listItem.measure(0, 0)
            val itemHeight = listItem.measuredHeight + binding.grepViewRecent.dividerHeight
            if (binding.grepViewRecent.height / itemHeight < titles.size) {
                binding.textViewShowAll.visibility = View.VISIBLE
            }
            binding.textViewRecentNotfound.visibility = View.GONE
        } else {
            binding.textViewRecentNotfound.visibility = View.VISIBLE
        }
        binding.progressBarRecent.visibility = View.GONE
        binding.grepViewRecent.visibility = View.VISIBLE
    }

    private fun updateReminders() = mainScope.launch(Dispatchers.Main) {
        binding.progressBarReminders.visibility = View.VISIBLE
        binding.grepViewReminders.visibility = View.INVISIBLE
        val reminders = grepReminders(applicationContext).toMutableList() // 削除できるようにmutable
        if (reminders.isNotEmpty()) {
            val adapter = GrepView.ReminderAdapter(this@HowmmMain, reminders)
            adapter.setFormat(null, howmmPrefs.fontSize)
            binding.grepViewReminders.adapter = adapter
            updateWidget(this@HowmmMain, adapter)
            binding.textViewRemindersNotfound.visibility = View.GONE
        } else {
            binding.textViewRemindersNotfound.visibility = View.VISIBLE
        }
        binding.progressBarReminders.visibility = View.GONE
        binding.grepViewReminders.visibility = View.VISIBLE
    }

    override fun dispatchKeyEvent(event: KeyEvent): Boolean {
        val keycode = event.keyCode
        if (event.action == KeyEvent.ACTION_UP && mViewerPopup.isShowing) {
            when (keycode) {
                KeyEvent.KEYCODE_SPACE -> {
                    if (event.isShiftPressed || event.isAltPressed) {
                        mViewerPopup.pageUp()
                    } else {
                        mViewerPopup.pageDown()
                    }
                    return true
                }
                KeyEvent.KEYCODE_PAGE_DOWN -> {
                    mViewerPopup.pageDown()
                    return true
                }
                KeyEvent.KEYCODE_PAGE_UP -> {
                    mViewerPopup.pageUp()
                    return true
                }
                else -> return super.dispatchKeyEvent(event)
            }
        }

        return super.dispatchKeyEvent(event)
    }

    override fun onPause() {
        if (isFinishing) {
            mainScope.cancel()
//            mReminderGrepTask?.cancel(true)
//            mTitleSearchTask?.cancel(true)
        }

        super.onPause()
    }

    override fun onDetachedFromWindow() {
        if (mViewerPopup.isShowing) { mViewerPopup.dismiss() }
        super.onDetachedFromWindow()
    }

    override fun onBackPressed() {
        if (mViewerPopup.isShowing) {
            mViewerPopup.dismiss()
        } else {
            super.onBackPressed()
        }
    }

    private fun editFile(item: GrepItem) {
        if (mViewerPopup.isShowing) { mViewerPopup.dismiss() }
        val intent = Intent(this@HowmmMain, TextEditor::class.java).apply {
            putExtra(TextEditor.KEY_PATHNAME, item.file.absolutePath)
            putExtra(TextEditor.KEY_LINENO, item.linenumber)
        }
        startActivity(intent)
    }

    private fun viewFile(item: GrepItem) {
        mViewerPopup.show(binding.linearLayoutMain)
        mViewerPopup.loadFile(null, item.file, item.linenumber)
    }

    private fun doSearch(query: String) {
        if (mViewerPopup.isShowing) { mViewerPopup.dismiss() }
        val intent = Intent(this@HowmmMain, HowmmResults::class.java).apply {
            action = Intent.ACTION_SEARCH
            putExtra(SearchManager.QUERY, query)
        }
        startActivity(intent)
    }

    private fun chooseHashtag() {
        val tags = howmmPrefs.getHashTags()
        val dialog = StringChoiceDialogFragment.newInstance(getString(R.string.label_hashtag), tags.toTypedArray())
        dialog.setListener(
                object : StringChoiceDialogFragment.Listener {
                    override fun onClick(result: String) = doSearch(result)
                })
        dialog.show(supportFragmentManager, "dialog")
    }

    private fun confirmDoneReminder(adapter: GrepView.ReminderAdapter, item: GrepItem) {
        val dialog = ConfirmationDialogFragment.newInstance(getString(R.string.message_confirm_done_reminder, item.text))
        dialog.setListener(
                object : ConfirmationDialogFragment.Listener {
                    override fun onPositiveClick() {
                        adapter.remove(item)
                        adapter.notifyDataSetChanged()
                        updateWidget(this@HowmmMain, adapter)
                        markReminderAsDone(item)
                    }

                    override fun onNegativeClick() {}
                })
        dialog.show(supportFragmentManager, "dialog")
    }

    private fun markReminderAsDone(item: GrepItem) {
        if (item.reminder == null) { return }

        val file = item.file
        val targetLine = item.linenumber

        val encoding = howmmPrefs.encoding
        val newline = howmmPrefs.newLine

        val lines = try {
            file.bufferedReader(charset(encoding)).use { it.readLines() }
        } catch (e: IOException) {
            Log.e("Howmm", "markReminderAsDone(): File reading error " + file.absolutePath)
            return
        }.toMutableList()

        lines[targetLine] = REGEX_REMINDER.replaceFirst(lines[targetLine], "$1.")

        try {
            file.bufferedWriter(charset(encoding)).use { bw ->
                lines.forEach {
                    bw.write(it)
                    bw.write(newline)
                }
            }
        } catch (e: IOException) {
            Log.e("Howmm", "markReminderAsDone(): File writing error " + file.absolutePath)
            return
        }
    }

    private fun updateWidget(context: Context, adapter: GrepView.ReminderAdapter) {
        Thread {
            val size = adapter.count
            val reminders = arrayOfNulls<String>(size)
            for (i in 0 until size) {
                val item = adapter.getItem(i)
                val rItem = item?.reminder
                rItem?.let {
                    reminders[i] =
                        item.text + "," + rItem.type.name + "," + rItem.priority.toString()
                }
            }

            val widgetIds = AppWidgetManager.getInstance(context)
                .getAppWidgetIds(ComponentName(context, HowmmReminderAppWidget::class.java))

            if (widgetIds.isNotEmpty()) {
                val update = Intent()
                update.putExtra(AppWidgetManager.EXTRA_APPWIDGET_IDS, widgetIds)
                update.putExtra(HowmmReminderAppWidget.EXTRA_HOWMM_WIDGET_SENDDATA, reminders)
                update.action = HowmmReminderAppWidget.ACTION_HOWMM_WIDGET_SENDDATA
                update.component = ComponentName(context, HowmmReminderAppWidget::class.java)
                context.sendBroadcast(update)
            }
        }.start()
    }
}
