package jp.deadend.noname.howmm

import java.io.File
import java.lang.IllegalArgumentException
import java.text.SimpleDateFormat
import java.util.Locale

import android.content.Context
import android.text.format.DateFormat
import android.util.Log
import androidx.preference.PreferenceManager

class Prefs(context: Context) {
    private val prefs = PreferenceManager.getDefaultSharedPreferences(context)
    private val res = context.resources
    private val defaultDir = context.getExternalFilesDir(null)?.absolutePath ?: ""

    var useRegex: Boolean
        get() = prefs.getBoolean(res.getString(R.string.pref_key_use_regex), false)
        set(value) = prefs.edit().putBoolean(res.getString(R.string.pref_key_use_regex), value).apply()

    var ignoreCase: Boolean
        get() = prefs.getBoolean(res.getString(R.string.pref_key_ignore_case), false)
        set(value) = prefs.edit().putBoolean(res.getString(R.string.pref_key_ignore_case), value).apply()

    var rootDir: String
        get() = prefs.getString(res.getString(R.string.pref_key_root_dir), null) ?: defaultDir
        set(value) = prefs.edit().putString(res.getString(R.string.pref_key_root_dir), value).apply()

    var extList: List<String>
        get() {
            val exts = prefs.getString(res.getString(R.string.pref_key_target_extensions), null) ?: res.getString(R.string.pref_value_target_extensions_default)
            return exts.dropLastWhile { it == '|' }.split("|")
        }
        set(value) = prefs.edit().
                putString(
                        res.getString(R.string.pref_key_target_extensions),
                        value.joinToString(separator = "|")
                ).apply()

    var encoding: String
        get() = prefs.getString(res.getString(R.string.pref_key_encoding), null) ?: res.getString(R.string.pref_value_encoding_default)
        set(value) = prefs.edit().putString(res.getString(R.string.pref_key_encoding), value).apply()

    var newLine: String
        get() {
            val key = res.getString(R.string.pref_key_newline)
            val defaultVal = res.getString(R.string.pref_value_newline_default)
            val str = prefs.getString(key, null) ?: defaultVal
            return try {
                NewLine.valueOf(str).value
            } catch (e: IllegalArgumentException) {
                prefs.edit().remove(key).apply()
                defaultVal
            }
        }
        set(value) = prefs.edit().putString(res.getString(R.string.pref_key_newline), value).apply()

    var fontSize: Int
        get() = prefs.getInt(res.getString(R.string.pref_key_font_size2), res.getInteger(R.integer.pref_value_font_size2_default))
        set(value) = prefs.edit().putString(res.getString(R.string.pref_key_font_size), value.toString()).apply()

    var reminderCompat: Boolean
        get() = prefs.getBoolean(res.getString(R.string.pref_key_reminder_compat), false)
        set(value) = prefs.edit().putBoolean(res.getString(R.string.pref_key_reminder_compat), value).apply()

    var defaultExt: String
        get() = prefs.getString(res.getString(R.string.pref_key_default_ext), null) ?: res.getString(R.string.pref_value_default_ext_default)
        set(value) = prefs.edit().putString(res.getString(R.string.pref_key_default_ext), value).apply()

    fun getShortcutKey(key: String) = prefs.getInt(key, SetKeyPreference.DEFAULT_VALUE)

    private fun getKeyWords(key: String): List<String> {
        val value = prefs.getString(key, "") ?: ""
        return if (value.isNotEmpty()) value.dropLastWhile { it == '|' }.split("|") else listOf()
    }

    fun getComeFromWords(): List<String> = getKeyWords(res.getString(R.string.pref_key_comefrom_words))
    fun getHashTags(): List<String> = getKeyWords(res.getString(R.string.pref_key_hashtags))

    private fun addKeyWords(key: String, words: List<String>) {
        val oldVal = prefs.getString(key, "")?.split("|")?.toMutableList() ?: mutableListOf()
        words.forEach { if (it !in oldVal) { oldVal.add(it) } }

        prefs.edit().putString(key, oldVal.joinToString(separator = "|")).apply()
    }

    fun addComeFromWords(words: List<String>) {
        addKeyWords(res.getString(R.string.pref_key_comefrom_words), words)
    }

    fun addHashTags(tags: List<String>) {
        addKeyWords(res.getString(R.string.pref_key_hashtags), tags)
    }

    fun getDateFormat(context: Context, with_time: Boolean): SimpleDateFormat {
        val dateFormatStr = prefs.getString(res.getString(R.string.pref_key_date_format), "") ?: ""
        if (dateFormatStr.isNotEmpty()) {
            val format = getFormat(dateFormatStr, with_time)
            if (format != null) { return format }
        }

        val systemFormat = (DateFormat.getDateFormat(context) as SimpleDateFormat).toPattern()
        val format = getFormat(systemFormat, with_time)
        if (format == null) {
            Log.e("Howmm", "Prefs#getDateFormat() bad system date format: $systemFormat")
            return DateFormat.getDateFormat(context) as SimpleDateFormat
        }

        return format
    }

    private fun getFormat(fmt_str: String?, with_time: Boolean): SimpleDateFormat? {
        val format = try {
            if (with_time) {
                SimpleDateFormat("[$fmt_str kk:mm]", Locale.US)
            } else {
                SimpleDateFormat("[$fmt_str]", Locale.US)
            }
        } catch (e: IllegalArgumentException) {
            null
        }

        return format
    }

    fun getDefaultFileList(): List<File> {
        val extList = extList.map { ".$it" } // ピリオドを付加
        return File(rootDir).walk().filter { file ->
            file.isFile && file.canRead() && extList.any { file.name.endsWith(it, ignoreCase = true) }
        }.toList()
    }
}

