package jp.deadend.noname.howmm

import java.text.SimpleDateFormat
import java.text.ParseException
import java.util.Calendar
import java.util.Date

import android.util.Log
import java.lang.NumberFormatException

class ReminderItem private constructor(
        private val mReminderDate: Date,
        val priority: Int,
        val type: ReminderType
) {
    val time: Long
        get() = mReminderDate.time

    enum class ReminderType(val defaultArg: Int) {
        APPT(1), DEADLINE(7), NOTE(1), SOMEDAY(30), DONE(1)
    }

    companion object {
        private const val ONE_DAY = 24L * 60L * 60L * 1000L
        private const val PRIORITY_BOTTOM = 65534
        private const val PRIORITY_INVISIBLE = 65535

        fun createInstance(
                item: GrepItem,
                dateFmt: SimpleDateFormat,
                dateTimefmt: SimpleDateFormat, // 引数の順番に依存しているので注意
                useOrigReminder: Boolean
        ): ReminderItem? {
            val hasTime: Boolean
            val text = item.text
            val date: Date
            Trace.d("text: $text")

            val d1 = try { dateFmt.parse(text) } catch (e: ParseException) { null }
            val d2 = try { dateTimefmt.parse(text) } catch (e: ParseException) { null }
            when {
                d1 != null -> {
                    date = d1
                    hasTime = false
                }
                d2 != null -> {
                    date = d2
                    hasTime = true
                }
                else -> {
                    Log.e("Howmm", "ReminderItem: bad date format: $text")
                    return null
                }
            }

            val matched = REGEX_REMINDER.find(text)
            val typeStr = matched?.groupValues?.get(2)
            var arg = -1
            matched?.groupValues?.get(3)?.let {
                try {
                    arg = it.toInt()
                } catch (e: NumberFormatException) {
                }
            }

            val type: ReminderType
            if (typeStr == null) {
                return null
            } else if (typeStr == "@") {
                type = ReminderType.APPT
                if (arg == -1) {
                    arg = ReminderType.APPT.defaultArg
                }
            } else if (typeStr == "!") {
                type = ReminderType.DEADLINE
                if (arg == -1) {
                    arg = if (useOrigReminder) ReminderType.DEADLINE.defaultArg else 1
                }
            } else if (typeStr == "-") {
                type = ReminderType.NOTE
                if (arg == -1) {
                    arg = ReminderType.NOTE.defaultArg
                }
            } else if (typeStr == "~") {
                type = ReminderType.SOMEDAY
                if (arg == -1) {
                    arg = ReminderType.SOMEDAY.defaultArg
                }
//            } else if (typeStr == ".") {
//                type = ReminderType.DONE
            } else {
                return null
            }

            val cal = Calendar.getInstance()
            cal.time = date
            Trace.d("cal: " + dateTimefmt.format(cal.time))

            val priority = if (useOrigReminder) {
                calcPriority_orig(type, arg, cal, hasTime)
            } else {
                calcPriority_new(type, arg, cal, hasTime)
            }
            if (priority == PRIORITY_INVISIBLE) { return null }

            if (!hasTime) {
                cal.add(Calendar.DAY_OF_MONTH, 1)
            } //時刻指定がない場合，表示順が最後になるように

            return ReminderItem(date, priority, type)
        }

        private fun calcPriority_orig(type: ReminderType, arg: Int, cal_start: Calendar, hasTime: Boolean): Int {
            val cal_now = Calendar.getInstance()
            val cal_end = cal_start.clone() as Calendar
            var priority: Int

            when (type) {
                ReminderType.APPT -> {
                    cal_end.set(Calendar.HOUR_OF_DAY, 0)
                    cal_end.set(Calendar.MINUTE, 0)
                    cal_end.set(Calendar.SECOND, 0)
                    cal_end.add(Calendar.DAY_OF_MONTH, arg)
                    if (cal_now.after(cal_end)) return PRIORITY_INVISIBLE // 過ぎてる

                    cal_start.set(Calendar.HOUR_OF_DAY, 0)
                    cal_start.set(Calendar.MINUTE, 0)
                    cal_start.set(Calendar.SECOND, 0)
                    cal_start.add(Calendar.DAY_OF_MONTH, 1)
                    val diff = cal_start.timeInMillis - cal_now.timeInMillis
                    priority = (diff / ONE_DAY).toInt()
                    if (priority < 0) { priority = 0 }
                }
                ReminderType.DEADLINE -> {
                    cal_start.add(Calendar.DAY_OF_MONTH, -arg)
                    if (cal_now.before(cal_start)) {
                        priority = PRIORITY_INVISIBLE // まだ非表示
                    } else {
                        if (!hasTime) {
                            cal_end.add(Calendar.DAY_OF_MONTH, 1)
                        } //時刻指定がない場合，その日の24時にセット
                        val diff = cal_end.timeInMillis - cal_now.timeInMillis
                        priority = (diff / ONE_DAY).toInt()
                        if (diff < 0) { priority-- } // 過ぎてる場合はマイナス1から
                    }
                }
                ReminderType.NOTE -> {
                    if (cal_now.before(cal_start)) {
                        priority = PRIORITY_INVISIBLE // まだ非表示
                    } else {
                        cal_end.add(Calendar.DAY_OF_MONTH, arg)
                        if (cal_now.before(cal_end)) {
                            priority = 0
                        } else {
                            val diff = cal_now.timeInMillis - cal_end.timeInMillis
                            priority = (diff / ONE_DAY).toInt() + 1
                            if (priority > PRIORITY_BOTTOM) { priority = PRIORITY_BOTTOM }
                        }
                    }
                }
                ReminderType.SOMEDAY -> {
                    cal_start.set(Calendar.HOUR_OF_DAY, 0)
                    cal_start.set(Calendar.MINUTE, 0)
                    cal_start.set(Calendar.SECOND, 0)

                    priority = if (cal_now.before(cal_start)) {
                        PRIORITY_INVISIBLE // まだ非表示
                    } else {
                        val diff = cal_now.timeInMillis - cal_start.timeInMillis
                        (diff / ONE_DAY).toInt() % arg
                    }
                    if (priority > arg / 2) { priority = arg - priority }
                }
                else -> return PRIORITY_INVISIBLE
            }

            Trace.d("priority: $priority")
            return priority
        }

        private fun calcPriority_new(type: ReminderType, arg: Int, cal_start: Calendar, hasTime: Boolean): Int {
            val cal_now = Calendar.getInstance()
            val cal_end = cal_start.clone() as Calendar
            var priority: Int

            when (type) {
                ReminderType.APPT -> {
                    cal_end.set(Calendar.HOUR_OF_DAY, 0)
                    cal_end.set(Calendar.MINUTE, 0)
                    cal_end.set(Calendar.SECOND, 0)
                    cal_end.add(Calendar.DAY_OF_MONTH, arg)
                    if (cal_now.after(cal_end)) return PRIORITY_INVISIBLE // 過ぎてる

                    cal_start.set(Calendar.HOUR_OF_DAY, 0)
                    cal_start.set(Calendar.MINUTE, 0)
                    cal_start.set(Calendar.SECOND, 0)
                    cal_start.add(Calendar.DAY_OF_MONTH, 1)
                    val diff = cal_start.timeInMillis - cal_now.timeInMillis

                    priority = (diff / ONE_DAY).toInt()
                    if (priority < 0) { priority = 0 }
                }
                ReminderType.DEADLINE -> {
                    if (!hasTime) {
                        cal_start.add(Calendar.DAY_OF_MONTH, 1)
                        cal_end.add(Calendar.DAY_OF_MONTH, 1)
                    } //時刻指定がない場合，その日の24時にセット
                    cal_start.add(Calendar.DAY_OF_MONTH, -arg)

                    priority = when {
                        cal_now.before(cal_start) -> {
                            val diff = cal_start.timeInMillis - cal_now.timeInMillis
                            (diff / ONE_DAY).toInt() + 1
                        }
                        cal_now.before(cal_end) -> 0 //オプション期間中はゼロ
                        else -> {
                            val diff = cal_end.timeInMillis - cal_now.timeInMillis
                            (diff / ONE_DAY).toInt() - 1
                        }
                    }
                }
                ReminderType.NOTE -> {
                    cal_end.add(Calendar.DAY_OF_MONTH, arg)

                    if (cal_now.before(cal_start)) {
                        val diff = cal_start.timeInMillis - cal_now.timeInMillis
                        priority = (diff / ONE_DAY).toInt() + 1
                    } else if (cal_now.before(cal_end)) {
                        priority = 0 //オプション期間中はゼロ
                    } else {
                        val diff = cal_now.timeInMillis - cal_end.timeInMillis
                        priority = (diff / ONE_DAY).toInt() + 1
                        if (priority > PRIORITY_INVISIBLE) { priority = PRIORITY_INVISIBLE }
                    }
                }
                ReminderType.SOMEDAY -> {
                    cal_start.set(Calendar.HOUR_OF_DAY, 0)
                    cal_start.set(Calendar.MINUTE, 0)
                    cal_start.set(Calendar.SECOND, 0)

                    priority = if (cal_now.before(cal_start)) {
                        val diff = cal_start.timeInMillis - cal_now.timeInMillis
                        ((diff / ONE_DAY).toInt() + 1) % arg
                    } else {
                        val diff = cal_now.timeInMillis - cal_start.timeInMillis
                        (diff / ONE_DAY).toInt() % arg
                    }
                    if (priority > arg / 2) { priority = arg - priority }
                }
                else -> return PRIORITY_INVISIBLE
            }

            Trace.d("priority: $priority")
            return priority
        }
    }
}
