package jp.deadend.noname.howmm

import android.app.Activity
import android.content.Intent
import android.os.Bundle
import android.view.Menu
import android.view.MenuItem
import androidx.activity.result.contract.ActivityResultContracts
import androidx.appcompat.app.AppCompatActivity
import androidx.preference.EditTextPreference
import androidx.preference.ListPreference
import androidx.preference.Preference
import androidx.preference.PreferenceFragmentCompat
import androidx.preference.PreferenceManager
import kotlinx.coroutines.*
import java.io.*
import java.lang.Exception
import java.text.SimpleDateFormat
import java.util.*
import java.util.zip.ZipEntry
import java.util.zip.ZipInputStream
import java.util.zip.ZipOutputStream

class SettingsActivity : AppCompatActivity() {
    class SettingsFragment : PreferenceFragmentCompat() {
        private val openDirActivity =
            registerForActivityResult(ActivityResultContracts.StartActivityForResult()) { result ->
                if (result.resultCode == Activity.RESULT_OK) {
                    result.data?.getStringExtra(FileChooser.KEY_FILEPATH)?.let {
                        howmmPrefs.rootDir = it
                        findPreference<Preference>(getString(R.string.pref_key_root_dir))?.summary = it
                    }
                }
            }

        override fun onCreatePreferences(savedInstanceState: Bundle?, rootKey: String?) {
            addPreferencesFromResource(R.xml.prefs)
            findPreference<Preference>(getString(R.string.pref_key_root_dir))
                    ?.summary = howmmPrefs.rootDir
            findPreference<EditTextPreference>(getString(R.string.pref_key_date_format))
                    ?.summaryProvider = EditTextPreference.SimpleSummaryProvider.getInstance()
            findPreference<EditTextPreference>(getString(R.string.pref_key_default_ext))
                    ?.summaryProvider = EditTextPreference.SimpleSummaryProvider.getInstance()
            findPreference<ListPreference>(getString(R.string.pref_key_encoding))
                    ?.summaryProvider = ListPreference.SimpleSummaryProvider.getInstance()
            findPreference<ListPreference>(getString(R.string.pref_key_newline))
                    ?.summaryProvider = ListPreference.SimpleSummaryProvider.getInstance()
        }

        override fun onDisplayPreferenceDialog(preference: Preference) {
            if (preference is SetKeyPreference) {
                val dialogFragment = SetKeyDialogFragment.newInstance(preference.key)
                dialogFragment.setTargetFragment(this, 0)
                dialogFragment.show(parentFragmentManager, "dialog")
            } else {
                super.onDisplayPreferenceDialog(preference)
            }
        }

        override fun onPreferenceTreeClick(preference: Preference): Boolean {
            if (preference.key == getString(R.string.pref_key_root_dir)) {
                val intent = Intent(requireContext(), FileChooser::class.java).apply {
                    putExtra(FileChooser.KEY_MODE, FileChooser.MODE_DIR)
                    putExtra(FileChooser.KEY_DIRNAME, howmmPrefs.rootDir)
                    putExtra(FileChooser.KEY_FONTSIZE, howmmPrefs.fontSize)
                }
                openDirActivity.launch(intent)
                return true
            }

            return super.onPreferenceTreeClick(preference)
        }
    }

    private lateinit var mProgressBar: MenuItem
    private val mainScope = MainScope()
    private val exportFileLauncher = registerForActivityResult(ActivityResultContracts.CreateDocument()) { uri ->
        if (uri != null) {
            try {
                contentResolver.openOutputStream(uri)?.use { zipTree(File(howmmPrefs.rootDir), it) }
            } catch (e: Exception) {
                SimpleMessageDialogFragment.newInstance(
                    getString(R.string.error_export, uri)
                ).show(supportFragmentManager, "dialog")
            }
        }
    }
    private val importFileLauncher = registerForActivityResult(ActivityResultContracts.OpenDocument()) { uri ->
        if (uri != null) {
            try {
                File(howmmPrefs.rootDir).listFiles()?.forEach { it.deleteRecursively() }
                contentResolver.openInputStream(uri)?.use { unzipFile(it, howmmPrefs.rootDir) }
            } catch (e: Exception) {
                SimpleMessageDialogFragment.newInstance(
                    getString(R.string.error_import, uri)
                ).show(supportFragmentManager, "dialog")
            }
        }
    }

    override fun onCreate(savedInstanceState: Bundle?) {
        super.onCreate(savedInstanceState)

        supportActionBar?.setDisplayHomeAsUpEnabled(true)
        title = getString(R.string.activity_title_preferences)

        supportFragmentManager.beginTransaction()
                .replace(android.R.id.content, SettingsFragment())
                .commit()
    }

    override fun onPause() {
        if (isFinishing) { mainScope.cancel() }
        super.onPause()
    }

    override fun onCreateOptionsMenu(menu: Menu): Boolean {
        menuInflater.inflate(R.menu.menu_settings, menu)
        mProgressBar = menu.findItem(R.id.menu_settings_progressbar)

        return super.onCreateOptionsMenu(menu)
    }

    override fun onOptionsItemSelected(item: MenuItem): Boolean {
        when (item.itemId) {
            R.id.menu_settings_rescan_keywords -> mainScope.launch(Dispatchers.Main) {
                mProgressBar.isVisible = true
                val comeFromWords = findComeFromWords()
                val hashTags = findHashTags()
                PreferenceManager.getDefaultSharedPreferences(applicationContext).edit()
                        .putString(resources.getString(R.string.pref_key_comefrom_words), comeFromWords)
                        .putString(resources.getString(R.string.pref_key_hashtags), hashTags)
                        .apply()

                mProgressBar.isVisible = false
                val dialog = SimpleMessageDialogFragment.newInstance(
                        getString(R.string.message_rescan_keywords,
                                comeFromWords.split("|").size, hashTags.split("|").size)
                )
                dialog.show(supportFragmentManager, "dialog")
            }
            R.id.menu_settings_clear_prefs -> {
                val dialog = ConfirmationDialogFragment.newInstance(getString(R.string.message_confirm_clear_prefs))
                dialog.setListener(
                        object : ConfirmationDialogFragment.Listener {
                            override fun onPositiveClick() {
                                PreferenceManager.getDefaultSharedPreferences(applicationContext)
                                        .edit()
                                        .clear()
                                        .apply()
                                supportFragmentManager.beginTransaction()
                                        .replace(android.R.id.content, SettingsFragment())
                                        .commit()
                            }

                            override fun onNegativeClick() {}
                        })
                dialog.show(supportFragmentManager, "dialog")
            }
            R.id.menu_settings_export -> {
                val dialog = ConfirmationDialogFragment.newInstance(getString(R.string.message_confirm_export))
                dialog.setListener(
                    object : ConfirmationDialogFragment.Listener {
                        override fun onPositiveClick() {
                            val sdf = SimpleDateFormat("'howmm-'yyyy-MM-dd-kkmmss'.zip'", Locale.US)
                            exportFileLauncher.launch(sdf.format(Date()))
                        }

                        override fun onNegativeClick() {}
                    })
                dialog.show(supportFragmentManager, "dialog")
            }
            R.id.menu_settings_import -> {
                val dialog = ConfirmationDialogFragment.newInstance(getString(R.string.message_confirm_import))
                dialog.setListener(
                    object : ConfirmationDialogFragment.Listener {
                        override fun onPositiveClick() = importFileLauncher.launch(arrayOf("*/*"))
                        override fun onNegativeClick() {}
                    })
                dialog.show(supportFragmentManager, "dialog")
            }
            android.R.id.home -> finish()
        }

        return super.onOptionsItemSelected(item)
    }

    private fun zipTree(file: File, target: OutputStream) {
        if (!file.isDirectory) { return }
        ZipOutputStream(target.buffered(BUFFER)).use { zipContents(it, file, file.path.length) }
    }
    private fun zipContents(zipOutput: ZipOutputStream, folder: File, basePathLength: Int) {
        val files = folder.listFiles() ?: return
        for (file in files) {
            if (file.isDirectory) {
                val entry = ZipEntry(file.path.substring(basePathLength).dropWhile { it == '/' } + "/")
                entry.time = file.lastModified()
                zipOutput.putNextEntry(entry)
                zipOutput.closeEntry()
                zipContents(zipOutput, file, basePathLength)
            } else {
                val entry = ZipEntry(file.path.substring(basePathLength).dropWhile { it == '/' })
                entry.time = file.lastModified()
                FileInputStream(file.path).buffered(BUFFER).use { input ->
                    zipOutput.putNextEntry(entry)
                    input.copyTo(zipOutput, BUFFER)
                    zipOutput.closeEntry()
                }
            }
        }
    }

    private fun unzipFile(src: InputStream, dstDir: String) {
        ZipInputStream(src.buffered(BUFFER)).use { input ->
            var zipEntry = input.nextEntry
            while (zipEntry != null) {
                val targetFile = File(dstDir, zipEntry.name.dropWhile { it == '/' }) // disable absolute paths
                if (zipEntry.isDirectory) {
                    targetFile.mkdirs()
                    targetFile.setLastModified(zipEntry.time)
                } else {
                    targetFile.parentFile?.let {
                        if (!it.exists()) { it.mkdirs() }
                    }
                    targetFile.outputStream().buffered(BUFFER).use { output ->
                        input.copyTo(output, BUFFER)
                    }
                    targetFile.setLastModified(zipEntry.time)
                }
                zipEntry = input.nextEntry
            }
        }
    }

    companion object {
        private const val BUFFER = 2048
    }
}