package jp.deadend.noname.howmm

import android.app.Activity
import java.text.SimpleDateFormat
import java.util.Calendar
import java.util.Date
import java.util.GregorianCalendar
import java.util.Locale

import android.app.Dialog
import android.content.Context
import android.content.Intent
import android.os.Build
import android.os.Bundle
import androidx.fragment.app.DialogFragment
import androidx.appcompat.app.AlertDialog
import androidx.appcompat.app.AppCompatActivity
import android.text.Editable
import android.text.TextWatcher
import android.view.inputmethod.InputMethodManager
import android.widget.AdapterView
import android.util.Log
import android.util.SparseIntArray
import android.view.*
import android.widget.EditText
import androidx.activity.result.contract.ActivityResultContracts
import jp.deadend.noname.howmm.databinding.ActivityEditorBinding
import jp.deadend.noname.howmm.databinding.FragmentDialogReminderBinding

import java.io.File
import java.io.IOException

class TextEditor : AppCompatActivity() {
    private lateinit var binding: ActivityEditorBinding
    private lateinit var editText: EditText
    private lateinit var mFile: File
    private var isModified = false
    private val mEditHistory = mutableListOf<EditHistoryItem>()
    private var isUndoing = false
    private val mShortcutKeys = SparseIntArray()
    private val getImageActivity = registerForActivityResult(ActivityResultContracts.GetContent()) { uri ->
        if (uri != null) {
            contentResolver.takePersistableUriPermission(uri, Intent.FLAG_GRANT_READ_URI_PERMISSION)
            val fileName = uri.path?.substringAfterLast('/') ?: ""
            insertStr(" ![$fileName]($uri) ")
        }
    }

    private data class EditHistoryItem(val mStart: Int, val mCount: Int, val mBeforeText: CharSequence)

    override fun onCreate(savedInstanceState: Bundle?) {
        super.onCreate(savedInstanceState)
        binding = ActivityEditorBinding.inflate(layoutInflater)
        setContentView(binding.root)
        editText = binding.editText
        supportActionBar?.setDisplayHomeAsUpEnabled(true)
        title = ""

        prepareText()

        editText.addTextChangedListener(object : TextWatcher {
            override fun beforeTextChanged(s: CharSequence, start: Int, count: Int, after: Int) {
                if (!isUndoing) {
                    val seq = s.subSequence(start, start + count)
                    mEditHistory.add(EditHistoryItem(start, after, seq))
                }
            }

            override fun onTextChanged(s: CharSequence, start: Int, before: Int, count: Int) {}
            override fun afterTextChanged(s: Editable) {
                isModified = true
            }
        })

        editText.textSize = howmmPrefs.fontSize.toFloat()
        setShortcutKeys()
    }

    override fun onCreateOptionsMenu(menu: Menu): Boolean {
        menuInflater.inflate(R.menu.menu_editor, menu)

        return super.onCreateOptionsMenu(menu)
    }

    override fun onOptionsItemSelected(item: MenuItem): Boolean {
        when (item.itemId) {
            R.id.menu_editor_undo -> performCommand(COMMAND_UNDO)
            R.id.menu_editor_date -> performCommand(COMMAND_DATE)
            R.id.menu_editor_hashtag -> chooseHashtag()
            R.id.menu_editor_image -> getImageActivity.launch("image/*")
            R.id.menu_editor_save_and_quit -> performCommand(COMMAND_DONE)
            R.id.menu_editor_remove_file -> performCommand(COMMAND_REMOVE_FILE)
            android.R.id.home -> performCommand(COMMAND_QUIT)
            else -> {}
        }

        return super.onOptionsItemSelected(item)
    }

    private fun prepareText() {
        val intent = intent
        if (intent != null) {
            val fname = intent.getStringExtra(KEY_PATHNAME)
            val dateTimeFormat = howmmPrefs.getDateFormat(applicationContext, true)
            if (fname == null) {
                val rootDir = howmmPrefs.rootDir

                val ext = howmmPrefs.defaultExt
                val sdf = SimpleDateFormat("yyyy/MM/yyyy-MM-dd-kkmmss.'$ext'", Locale.US)
                mFile = File(rootDir + "/" + sdf.format(Date()))

                val title = intent.getStringExtra(KEY_TITLE) ?: ""
                editText.setText("= $title\n\n${dateTimeFormat.format(Date())}")
                editText.setSelection(2)
            } else {
                mFile = File(fname)
                if (!mFile.exists()) {
                    editText.setText("= $title\n\n${dateTimeFormat.format(Date())}")
                    editText.setSelection(2)
                } else if (mFile.isFile && mFile.canRead()) {
                    val encoding = howmmPrefs.encoding
                    val str = try {
                        mFile.bufferedReader(charset(encoding)).use { it.readText() }
                    } catch (e: IOException) {
                        Log.e("Howmm", "prepareText(): File IO error " + mFile.absolutePath)
                        ""
                    }
                    editText.setText(str)
                    val line = intent.getIntExtra(KEY_LINENO, 0)
                    editText.setSelection(str.getOffsetForLine(line))
                } else {
                    setResult(Activity.RESULT_CANCELED)
                    finish()
                }
            }
        } else {
            setResult(Activity.RESULT_CANCELED)
            finish()
        }
    }

    private fun chooseHashtag() {
        val tags = howmmPrefs.getHashTags()
        val dialog = StringChoiceDialogFragment.newInstance(getString(R.string.label_hashtag), tags.toTypedArray())
        dialog.setListener(object : StringChoiceDialogFragment.Listener {
            override fun onClick(result: String) { insertStr(result) }
        })
        dialog.show(supportFragmentManager, "dialog")
    }

    private fun setShortcutKeys() {
        mShortcutKeys.put(
                howmmPrefs.getShortcutKey(getString(R.string.pref_key_shortcut_date)), COMMAND_DATE)
        mShortcutKeys.put(
                howmmPrefs.getShortcutKey(getString(R.string.pref_key_shortcut_curdatetime)), COMMAND_CUR_DATETIME)
        mShortcutKeys.put(
                howmmPrefs.getShortcutKey(getString(R.string.pref_key_shortcut_undo)), COMMAND_UNDO)
        mShortcutKeys.put(
                howmmPrefs.getShortcutKey(getString(R.string.pref_key_shortcut_done)), COMMAND_DONE)

        binding.MainLayout.setCallback(object : LinearLayoutKeyDispatcher.Callback {
            override fun commandShortcut(key: Int): Boolean {
                val command = mShortcutKeys.get(key)
                if (command == 0) { return false }

                performCommand(command)
                return true
            }
        })
    }

    override fun dispatchKeyEvent(event: KeyEvent): Boolean {
        if (event.action == KeyEvent.ACTION_DOWN) {
            when (event.keyCode) {
                KeyEvent.KEYCODE_ENTER ->
                    //enterキーリピート防止
                    if (event.repeatCount > 0) { return true }
                KeyEvent.KEYCODE_BACK -> {
                    performCommand(COMMAND_QUIT)
                    return true
                }
                KeyEvent.KEYCODE_SEARCH -> {
                    performCommand(COMMAND_SEARCH)
                    return true
                }
            }
        }

        return super.dispatchKeyEvent(event)
    }

    private fun performCommand(cmd: Int) {
        when (cmd) {
            COMMAND_DATE -> ReminderDialogFragment().show(supportFragmentManager, "datePicker")
            COMMAND_CUR_DATETIME -> {
                val dateTimeFormat = howmmPrefs.getDateFormat(applicationContext, true)
                insertStr(dateTimeFormat.format(Date()))
            }
            COMMAND_UNDO -> undo()
            COMMAND_SEARCH -> { }
            COMMAND_DONE -> processDone()
            COMMAND_QUIT -> if (isModified) {
                val dialog = ConfirmationDialogFragment.newInstance(getString(R.string.message_confirm_exit))
                dialog.setListener(
                        object : ConfirmationDialogFragment.Listener {
                            override fun onPositiveClick() {
                                setResult(Activity.RESULT_CANCELED)
                                finish()
                            }

                            override fun onNegativeClick() {}
                        })
                dialog.show(supportFragmentManager, "dialog")
            } else {
                setResult(Activity.RESULT_CANCELED)
                finish()
            }
            COMMAND_REMOVE_FILE -> removeFile()
            else -> { }
        }
    }

    private fun removeFile() {
        val dialog = ConfirmationDialogFragment.newInstance(getString(R.string.message_confirm_remove_file, mFile.toString()))
        dialog.setListener(
                object : ConfirmationDialogFragment.Listener {
                    override fun onPositiveClick() {
                        mFile.delete()
                        setResult(Activity.RESULT_OK)
                        finish()
                    }

                    override fun onNegativeClick() {}
                })
        dialog.show(supportFragmentManager, "dialog")
    }

    private fun processDone() {
        when {
            editText.text.isEmpty() -> removeFile()
            isModified -> {
                saveToFile()

                val comeFromWords = REGEX_COMEFROM_LINK.findAll(editText.text)
                        .map { it.groupValues[1] }.distinct().toList()
                val hashTags = REGEX_HASHTAG.findAll(editText.text)
                        .map { it.groupValues[it.groupValues.size - 1] }.distinct().toList()
                howmmPrefs.addComeFromWords(comeFromWords)
                howmmPrefs.addHashTags(hashTags)

                setResult(Activity.RESULT_OK)
                finish()
            }
            else -> {
                setResult(Activity.RESULT_CANCELED)
                finish()
            }
        }
    }

    private fun insertStr(str: CharSequence) {
        val pos = editText.selectionStart
        editText.text.insert(pos, str)
    }

    private fun undo() {
        if (mEditHistory.isEmpty()) { return }

        val item = mEditHistory.last()
        isUndoing = true
        editText.text.replace(item.mStart, item.mStart + item.mCount, item.mBeforeText)
        isUndoing = false
        mEditHistory.removeAt(mEditHistory.size - 1)
    }

    private fun saveToFile(): Boolean {
        if (!mFile.exists()) { mFile.parentFile?.mkdirs() }

        val encoding = howmmPrefs.encoding
        val newline = howmmPrefs.newLine
        try {
            mFile.bufferedWriter(charset(encoding)).use {
                it.write(editText.text.toString().replace("\n", newline))
            }
        } catch (e: Exception) {
            Log.e("Howmm", "saveToFile(): File IO error")
            return false
        }

        return true
    }

    override fun finish() {
        val imm = getSystemService(Context.INPUT_METHOD_SERVICE) as InputMethodManager
        imm.hideSoftInputFromWindow(editText.windowToken, 0)
        super.finish()
    }

    class ReminderDialogFragment : DialogFragment() {
        private var mType = 0

        override fun onCreateDialog(savedInstanceState: Bundle?): Dialog {
            val inflater = requireContext().getSystemService(Context.LAYOUT_INFLATER_SERVICE) as LayoutInflater
            val binding = FragmentDialogReminderBinding.inflate(inflater)
            val lv = binding.listViewType
            val dp = binding.datePicker
            val tp = binding.timePicker
            val move_to_tp = binding.buttonTime
            val dateFormat = howmmPrefs.getDateFormat(requireContext(), false)
            val dateTimeFormat = howmmPrefs.getDateFormat(requireContext(), true)

            lv.onItemClickListener = AdapterView.OnItemClickListener { _, _, position, _ ->
                mType = position
                lv.visibility = View.GONE
                dp.visibility = View.VISIBLE
                move_to_tp.visibility = View.VISIBLE
            }

            tp.setIs24HourView(true)
            move_to_tp.setOnClickListener {
                dp.visibility = View.GONE
                move_to_tp.visibility = View.GONE
                tp.visibility = View.VISIBLE
            }

            return AlertDialog.Builder(requireContext())
                    .setView(binding.root)
                    .setPositiveButton(android.R.string.ok) { _, _ ->
                        val mark = when (mType) {
                            0 -> "@ "
                            1 -> "! "
                            2 -> "- "
                            3 -> "~ "
                            4 -> ""
                            else -> ""
                        }
                        val cal: Calendar
                        val activity = activity as TextEditor
                        if (tp.isShown) {
                            cal = if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.M) {
                                GregorianCalendar(dp.year, dp.month, dp.dayOfMonth, tp.hour, tp.minute)
                            } else {
                                GregorianCalendar(dp.year, dp.month, dp.dayOfMonth, tp.currentHour, tp.currentMinute)
                            }
                            activity.insertStr(dateTimeFormat.format(cal.getTime()) + mark)
                        } else {
                            cal = GregorianCalendar(dp.year, dp.month, dp.dayOfMonth)
                            activity.insertStr(dateFormat.format(cal.getTime()) + mark)

                        }
                        dismiss()
                    }
                    .setNegativeButton(android.R.string.cancel) { _, _ -> dismiss() }
                    .create()
        }
    }

    companion object {
        const val KEY_PATHNAME = "TextEditorPathName"
        const val KEY_LINENO = "TextEditorLineNumber"
        const val KEY_TITLE = "TextEditorTitle"

        private const val COMMAND_QUIT = 1
        private const val COMMAND_DONE = 2
        private const val COMMAND_SEARCH = 3
        private const val COMMAND_DATE = 4
        private const val COMMAND_CUR_DATETIME = 5
        private const val COMMAND_UNDO = 6
        private const val COMMAND_REMOVE_FILE = 7
    }
}