package jp.deadend.noname.howmm

import java.io.File

import android.app.SearchManager
import android.content.Context
import android.content.Intent
import android.graphics.BitmapFactory
import android.graphics.Typeface
import android.net.Uri
import android.text.Spannable
import android.text.SpannableString
import android.text.SpannableStringBuilder
import android.text.style.*
import android.view.View
import android.webkit.MimeTypeMap
import java.util.*

object TextSpanner {
    private const val HIGHLIGHT_BG = -0xff0001
    private const val COLOR_TITLE = -0xffff6a
    private const val COLOR_COMEFROM = -0xff6a00
    private const val COLOR_REMINDER_TODAY_BG = -0x2b00
    private const val COLOR_REMINDER_PASSED_BG = -0x8210
    private val REMINDER_COLOR_MAP =
            EnumMap<ReminderItem.ReminderType, Int>(ReminderItem.ReminderType::class.java).apply {
                put(ReminderItem.ReminderType.APPT,     -0x1000000)
                put(ReminderItem.ReminderType.DEADLINE, -0x6a0000)
                put(ReminderItem.ReminderType.NOTE,     -0xffff01)
                put(ReminderItem.ReminderType.SOMEDAY,  -0xff8800)
            }

    // markdown
    private const val PATTERN_CODE_MULTILINE = "```"
    private val REGEX_LISTITEM =         Regex("""^( *)[-+*] (.*)$""")
    private val REGEX_ORDERED_LISTITEM = Regex("""^( *)([0-9]+\.) (.*)$""")
    private val REGEX_BLOCKQUOTE =       Regex("""^> (.*)$""")
    private val REGEX_EMPH =             Regex("""[^*_]([*_])([^*_]+?)(\1)[^*_]""")
    private val REGEX_STRONG =           Regex("""[^*_](\*\*|__)([^*_]+?)(\1)[^*_]""")
    private val REGEX_CODE =             Regex("""(`)([^`]+)(`)""")
    private val REGEX_IMAGE =            Regex("""!\[.+]\((.+)\)""")

    fun spanGrepAdapter(text: String, regex: Regex?): SpannableString {
        val ss = SpannableString(text)

        if (Regex(REGEX_PATTERN_TITLE).containsMatchIn(text)) { // title line
            val fgspan = ForegroundColorSpan(COLOR_TITLE)
            ss.setSpan(fgspan, 0, ss.length, Spannable.SPAN_EXCLUSIVE_EXCLUSIVE)
        }

        REGEX_COMEFROM_LINK.find(text)?.let { // come-from
            val fgspan = ForegroundColorSpan(COLOR_COMEFROM)
            ss.setSpan(fgspan, it.range.first, it.range.last + 1, Spannable.SPAN_EXCLUSIVE_EXCLUSIVE)
        }

        regex?.findAll(text)?.forEach { // matched words
            val bgspan = BackgroundColorSpan(HIGHLIGHT_BG)
            ss.setSpan(bgspan, it.range.first, it.range.last + 1, Spannable.SPAN_EXCLUSIVE_EXCLUSIVE)
        }

        return ss
    }

    fun spanReminderAdapter(text: CharSequence, type: ReminderItem.ReminderType, priority: Int): SpannableString {
        val ss = SpannableString(text)
        val fgColor = REMINDER_COLOR_MAP[type] ?: return ss

        ss.setSpan(ForegroundColorSpan(fgColor), 0, ss.length, Spannable.SPAN_EXCLUSIVE_EXCLUSIVE)

        if (priority == 0) {
            ss.setSpan(
                    BackgroundColorSpan(COLOR_REMINDER_TODAY_BG),
                    0, ss.length, Spannable.SPAN_EXCLUSIVE_EXCLUSIVE
            )
        } else if (priority < 0) {
            ss.setSpan(
                    BackgroundColorSpan(COLOR_REMINDER_PASSED_BG),
                    0, ss.length, Spannable.SPAN_EXCLUSIVE_EXCLUSIVE
            )
        }

        return ss
    }

    class SearchLinkSpan(private val mString: String) : ClickableSpan() {
        override fun onClick(v: View) {
            val context = v.context
            val intent: Intent
            val f = File(mString)
            if (f.canRead()) {
                val mimeType = MimeTypeMap.getFileExtensionFromUrl(f.name)
                        ?.let { MimeTypeMap.getSingleton().getMimeTypeFromExtension(it.lowercase(Locale.US)) }
                        ?: "*/*"
                intent = Intent(Intent.ACTION_VIEW).apply {
                    setDataAndType(Uri.fromFile(f), mimeType)
                    putExtra(Intent.EXTRA_STREAM, Uri.fromFile(f))
                }
            } else {
                intent = Intent(context, HowmmResults::class.java).apply {
                    action = Intent.ACTION_SEARCH
                    flags = Intent.FLAG_ACTIVITY_CLEAR_TOP
                    putExtra(SearchManager.QUERY, mString)
                }
            }
            context.startActivity(intent)
        }
    }

    private fun spanMarkDown(text: List<String>, context: Context): SpannableStringBuilder {
        val density = context.resources.displayMetrics.density
        val ssb = SpannableStringBuilder()

        var isMultiLineCode = false
        text.forEach { line ->
            val spanStart = ssb.length
            when {
                line.startsWith(PATTERN_CODE_MULTILINE) -> { // Code block marks
                    isMultiLineCode = !isMultiLineCode
                    return@forEach
                }
                isMultiLineCode -> { // Code blocks
                    ssb.append("$line\n")
                    if (line.isNotEmpty()) {
                        ssb.setSpan(TypefaceSpan("monospace"), spanStart, ssb.length, Spannable.SPAN_EXCLUSIVE_EXCLUSIVE)
                        ssb.setSpan(RelativeSizeSpan(0.8f), spanStart, ssb.length, Spannable.SPAN_EXCLUSIVE_EXCLUSIVE)
                    }
                    return@forEach
                }
            }

            val size = howmmPrefs.fontSize
            REGEX_LISTITEM.find(line)?.let { // Unordered lists
                val item = it.groupValues[2]
                val indentWidth = it.groupValues[1].length
                ssb.append("$item\n")
                ssb.setSpan(ItemSpan((size / 4 + size * indentWidth / 2) * density, size * density / 6), spanStart, ssb.length, Spannable.SPAN_EXCLUSIVE_EXCLUSIVE)
                return@forEach
            }
            REGEX_ORDERED_LISTITEM.find(line)?.let { // Ordered lists
                val item = it.groupValues[3]
                val indentWidth = it.groupValues[1].length
                ssb.append("$item\n")
                ssb.setSpan(HeadedSpan(it.groupValues[2], (size / 4 + size * indentWidth / 2) * density, size * density / 6), spanStart, ssb.length, Spannable.SPAN_EXCLUSIVE_EXCLUSIVE)
                return@forEach
            }
            REGEX_BLOCKQUOTE.find(line)?.let { // Blockquotes
                val item = it.groupValues[1]
                ssb.append("$item\n")
                ssb.setSpan(LeadingMarginSpan.Standard((size * density).toInt()), spanStart, ssb.length, Spannable.SPAN_EXCLUSIVE_EXCLUSIVE)
                ssb.setSpan(QuoteSpan(), spanStart, ssb.length, Spannable.SPAN_EXCLUSIVE_EXCLUSIVE)
                return@forEach
            }

            ssb.append("$line\n") // others
        }

        REGEX_CODE.findAll(ssb).forEach {
            val groups = it.groups
            ssb.setSpan(TypefaceSpan("monospace"), groups[2]!!.range.first, groups[2]!!.range.last + 1, Spannable.SPAN_EXCLUSIVE_EXCLUSIVE)
            ssb.setSpan(RelativeSizeSpan(0.8f), groups[2]!!.range.first, groups[2]!!.range.last + 1, Spannable.SPAN_EXCLUSIVE_EXCLUSIVE)
            ssb.setSpan(NoDisplaySpan(), groups[1]!!.range.first, groups[1]!!.range.last + 1, Spannable.SPAN_EXCLUSIVE_EXCLUSIVE)
            ssb.setSpan(NoDisplaySpan(), groups[3]!!.range.first, groups[3]!!.range.last + 1, Spannable.SPAN_EXCLUSIVE_EXCLUSIVE)
        }

        REGEX_EMPH.findAll(ssb).forEach {
            val groups = it.groups
            val spans1 = ssb.getSpans(groups[2]!!.range.first, groups[2]!!.range.last + 1, TypefaceSpan::class.java)
            val spans2 = ssb.getSpans(groups[2]!!.range.first, groups[2]!!.range.last + 1, QuoteSpan::class.java)
            if (spans1.isEmpty() && spans2.isEmpty()) {
                ssb.setSpan(FakeBoldItalicSpan(Typeface.ITALIC), groups[2]!!.range.first, groups[2]!!.range.last + 1, Spannable.SPAN_EXCLUSIVE_EXCLUSIVE)
                ssb.setSpan(NoDisplaySpan(), groups[1]!!.range.first, groups[1]!!.range.last + 1, Spannable.SPAN_EXCLUSIVE_EXCLUSIVE)
                ssb.setSpan(NoDisplaySpan(), groups[3]!!.range.first, groups[3]!!.range.last + 1, Spannable.SPAN_EXCLUSIVE_EXCLUSIVE)
            }
        }

        REGEX_STRONG.findAll(ssb).forEach {
            val groups = it.groups
            val spans1 = ssb.getSpans(groups[2]!!.range.first, groups[2]!!.range.last + 1, TypefaceSpan::class.java)
            val spans2 = ssb.getSpans(groups[2]!!.range.first, groups[2]!!.range.last + 1, QuoteSpan::class.java)
            if (spans1.isEmpty() && spans2.isEmpty()) {
                ssb.setSpan(FakeBoldItalicSpan(Typeface.BOLD), groups[2]!!.range.first, groups[2]!!.range.last + 1, Spannable.SPAN_EXCLUSIVE_EXCLUSIVE)
                ssb.setSpan(NoDisplaySpan(), groups[1]!!.range.first, groups[1]!!.range.last + 1, Spannable.SPAN_EXCLUSIVE_EXCLUSIVE)
                ssb.setSpan(NoDisplaySpan(), groups[3]!!.range.first, groups[3]!!.range.last + 1, Spannable.SPAN_EXCLUSIVE_EXCLUSIVE)
            }
        }

        REGEX_IMAGE.findAll(ssb).forEach {
            val uriStr = it.groupValues[1]
            if (uriStr.startsWith("content:")) {
                val imageSpan = ImageSpan(context, Uri.parse(it.groupValues[1]))
                if (imageSpan.drawable != null) {
                    ssb.setSpan(imageSpan, it.range.first, it.range.last + 1, Spannable.SPAN_EXCLUSIVE_EXCLUSIVE)
                }
            } else {
                val bitmap = BitmapFactory.decodeFile(uriStr)
                if (bitmap != null) {
                    ssb.setSpan(ImageSpan(context, bitmap), it.range.first, it.range.last + 1, Spannable.SPAN_EXCLUSIVE_EXCLUSIVE)
                }
            }
        }

        return ssb
    }

    fun spanTextViewer(text: List<String>, regex: Regex?, context: Context): SpannableStringBuilder {
        val ssb = spanMarkDown(text, context)

        Regex(REGEX_PATTERN_TITLE, RegexOption.MULTILINE).findAll(ssb).forEach { // title lines
            val fgspan = ForegroundColorSpan(COLOR_TITLE)
            ssb.setSpan(fgspan, it.range.first, it.range.last + 1, Spannable.SPAN_EXCLUSIVE_EXCLUSIVE)
        }

        REGEX_HASHTAG.findAll(ssb).forEach { // hashtags
            val group = it.groupValues
            val slspan = SearchLinkSpan(group[group.size - 1])
            ssb.setSpan(slspan, it.range.first, it.range.last + 1, Spannable.SPAN_EXCLUSIVE_EXCLUSIVE)
        }

        regex?.findAll(ssb)?.forEach { // matched words
            val bgspan = BackgroundColorSpan(HIGHLIGHT_BG)
            ssb.setSpan(bgspan, it.range.first, it.range.last + 1, Spannable.SPAN_EXCLUSIVE_EXCLUSIVE)
        }

        REGEX_GOTO_LINK.findAll(ssb).forEach { // goto links
            val group = it.groupValues
            val slspan = SearchLinkSpan(group[group.size - 1])
            ssb.setSpan(slspan, it.range.first, it.range.last + 1, Spannable.SPAN_EXCLUSIVE_EXCLUSIVE)
            ssb.getSpans(it.range.first, it.range.last + 1, NoDisplaySpan::class.java)
                    .forEach { span -> ssb.removeSpan(span) }
            ssb.getSpans(it.range.first, it.range.last + 1, FakeBoldItalicSpan::class.java)
                    .forEach { span -> ssb.removeSpan(span) }
        }

        val comeFromWords = howmmPrefs.getComeFromWords()
        for (word in comeFromWords) { // come-from links
            Regex(word).findAll(ssb).forEach {
                val slspan = SearchLinkSpan(word)
                ssb.setSpan(slspan, it.range.first, it.range.last + 1, Spannable.SPAN_EXCLUSIVE_EXCLUSIVE)
            }
        }

        return ssb
    }
}
