package jp.deadend.noname.howmm

import android.content.Context
import kotlinx.coroutines.*
import java.io.File
import java.io.IOException

const val REGEX_PATTERN_TITLE = """^\s*=.*"""
val REGEX_REMINDER      = Regex("""(^\s*\[.+]\s*)([@!\-+~])(\d*)""")
val REGEX_HASHTAG       = Regex("""(?:^|\s|[\p{Punct}&&[^/]])(#[\S&&[^#]]+)""")
val REGEX_COMEFROM_LINK = Regex("""<<<\s*(\S+)""")
val REGEX_GOTO_LINK     = Regex(""">>>\s*(\S+)""")

private fun grep(cacheFileName: String?, regex: Regex): List<GrepItem> {
    val encoding = charset(howmmPrefs.encoding)

    fun grepFile(file: File): List<GrepItem> {
        val lines = try {
            file.bufferedReader(encoding).readLines()
        } catch (e: IOException) {
            return listOf()
        }

        val result = mutableListOf<GrepItem>()
        lines.forEachIndexed { index, s ->
            if (regex.containsMatchIn(s)) result.add(GrepItem(file, index, s))
        }

        return result
    }

    val cacheFile: File?
    val result = mutableListOf<GrepItem>()
    val filesList: List<File>

    if (cacheFileName != null) {
        cacheFile = File(howmmPrefs.rootDir, cacheFileName)
        val (cachedItems, cacheLastModified) = loadCacheFile(cacheFile, encoding)
        result.addAll(cachedItems)

        val allFilesList = howmmPrefs.getDefaultFileList()
        filesList = allFilesList.filter { it.lastModified() >= cacheLastModified }
    } else {
        cacheFile = null
        filesList = howmmPrefs.getDefaultFileList()
    }

    runBlocking {
        result.addAll(
            filesList.map { async(Dispatchers.Default) { grepFile(it) } }.flatMap { it.await() }
        )
    }

    if (cacheFile != null && result.isNotEmpty()) {
        saveCacheFile(cacheFile, encoding, howmmPrefs.newLine, result)
    }

    return result
}

suspend fun grepReminders(context: Context): List<GrepItem> = withContext(Dispatchers.Default) {
    val tmpResult = grep(REMINDER_CACHE_FILE, REGEX_REMINDER)

    val dateFormat = howmmPrefs.getDateFormat(context, false)
    val dateTimeFormat = howmmPrefs.getDateFormat(context, true)
    val reminderCompat = howmmPrefs.reminderCompat
    tmpResult.forEach { it.updateReminder(dateFormat, dateTimeFormat, reminderCompat) }

    tmpResult.filter { it.reminder != null }
            .sortedWith(compareBy({ it.reminder?.priority }, { it.text }))
}

suspend fun grepTitles(): List<GrepItem> = withContext(Dispatchers.Default) {
    grep(TITLE_CACHE_FILE, Regex(REGEX_PATTERN_TITLE))
}

suspend fun grepRegex(regex: Regex): List<GrepItem> = withContext(Dispatchers.Default) {
    grep(null, regex)
}

suspend fun grepMultipleWords(
        words: List<String>,
        ignoreCase: Boolean
): List<GrepItem> = withContext(Dispatchers.Default) {
    val encoding = charset(howmmPrefs.encoding)

    fun grepFile(file: File): List<GrepItem> {
        val foundFlags = MutableList(words.size) { false }
        val lines = try {
            file.bufferedReader(encoding).readLines()
        } catch (e: IOException) {
            return listOf()
        }

        val result = mutableListOf<GrepItem>()
        lines.forEachIndexed { index, s ->
            words.forEachIndexed { indexOfWord, word ->
                if (s.indexOf(word, ignoreCase = ignoreCase) != -1) {
                    foundFlags[indexOfWord] = true
                    result.add(GrepItem(file, index, s))
                }
            }
        }

        return if (foundFlags.all {it}) result.distinct() else listOf()
    }

    howmmPrefs.getDefaultFileList()
            .map { async { grepFile(it) } }
            .flatMap { it.await() }
}

suspend fun grepHashtags(hashtag: String): List<GrepItem> = withContext(Dispatchers.Default) {
    val encoding = charset(howmmPrefs.encoding)
    val regexTitle = Regex(REGEX_PATTERN_TITLE)

    fun grepFile(file: File): List<GrepItem> {
        val lines = try {
            file.bufferedReader(encoding).readLines()
        } catch (e: IOException) {
            return listOf()
        }

        var titleLine = GrepItem(file, 0, lines[0])
        val result = mutableListOf<GrepItem>()
        lines.forEachIndexed { index, s ->
            when {
                s.indexOf(hashtag) != -1 -> result.add(titleLine)
                regexTitle.containsMatchIn(s) -> titleLine = GrepItem(file, index, s)
            }
        }

        return result
    }

    howmmPrefs.getDefaultFileList()
            .map { async { grepFile(it) } }
            .flatMap { it.await() }
}

private fun findWords(regex: Regex, getWord: (MatchResult) -> String): String {
    val encoding = charset(howmmPrefs.encoding)

    fun searchFile(file: File): List<String> {
        val lines = try {
            file.bufferedReader(encoding).readLines()
        } catch (e: IOException) {
            return listOf()
        }

        val result = mutableListOf<String>()
        lines.forEach { s ->
            val matched = regex.findAll(s).map(getWord)
            result.addAll(matched)
        }

        return result
    }

    return runBlocking {
        howmmPrefs.getDefaultFileList()
                .map { async(Dispatchers.Default) {searchFile(it)} }
                .flatMap { it.await() }
                .distinct().joinToString(separator = "|")
    }
}

suspend fun findHashTags(): String = withContext(Dispatchers.Default) {
    findWords(REGEX_HASHTAG) { result ->
        val group = result.groupValues
        group[group.size - 1]
    }
}

suspend fun findComeFromWords(): String = withContext(Dispatchers.Default) {
    findWords(REGEX_COMEFROM_LINK) { it.destructured.component1() }
}
